;; Generated package description from lin.el  -*- no-byte-compile: t -*-
(define-package "lin" "1.1.0.0.20260211.85534" "Make `hl-line-mode' more suitable for selection UIs" '((emacs "27.1")) :commit "d679b80c4a42dd3ee8c00ab87b4bd3414781b1d3" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "faces" "theme") :url "https://github.com/protesilaos/lin")
